# Standard Library
import logging
from typing import Optional

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status
from sqlalchemy.exc import IntegrityError

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, RWS_ROUTE_VCS, VCS_TAG
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import vcs_instance as vcs_instance_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema import vcs_instance as vcs_instance_schema
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(prefix=f"{RWS_ROUTE_VCS}", tags=[VCS_TAG])
logger = logging.getLogger(__name__)


@router.post("", response_model=vcs_instance_schema.VCSInstanceRead, status_code=status.HTTP_201_CREATED)
def create_vcs_instance(vcs_instance: vcs_instance_schema.VCSInstanceCreate,
                        db_connection: Session = Depends(get_db_connection)) \
        -> vcs_instance_schema.VCSInstanceRead:
    """
        Create new VCS instance object
    :param db_connection:
        Session of the database connection
    :param vcs_instance:
        VCSInstanceCreate type object of the VCS Instance to create
    :return: VCSInstanceRead
        The output will contain a VCSInstanceRead type object if the creation was successful
    """
    try:
        created_vcs_instance = vcs_instance_crud.create_vcs_instance_if_not_exists(db_connection=db_connection,
                                                                                   vcs_instance=vcs_instance)
    except IntegrityError as err:
        logger.debug(f"Error creating new vcs_instance object: {err}")
        raise HTTPException(status_code=400, detail="Error creating new vcs_instance object") from err
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(created_vcs_instance)
    return vcs_instance


@router.get("/{vcs_instance_id}", response_model=vcs_instance_schema.VCSInstanceRead, status_code=status.HTTP_200_OK)
def read_vcs_instance(vcs_instance_id: int, db_connection: Session = Depends(get_db_connection)) \
        -> vcs_instance_schema.VCSInstanceRead:
    """
        Get a VCS instance object based on the provided id
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the VCS instance to get
    :return: VCSInstanceRead
        The output will contain a VCSInstanceRead type object from the requested id
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS Instance not found")
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance)
    return vcs_instance


@router.get("", response_model=PaginationModel[vcs_instance_schema.VCSInstanceRead], status_code=status.HTTP_200_OK)
def get_all_vcs_instances(skip: int = Query(default=0, ge=0),
                          limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                          vcs_provider_type: Optional[VCSProviders] = Query(None),
                          vcs_instance_name: Optional[str] = Query(None),
                          db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[vcs_instance_schema.VCSInstanceRead]:
    """
        Retrieve all VCS instance objects paginated
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param vcs_provider_type:
        Optional filter for the provider type
    :param vcs_instance_name:
        Optional filter for the provider name
    :return: [VCSInstanceRead]
        The output will contain a PaginationModel containing the list of VCSInstanceRead type objects,
        or an empty list if no VCS instance was found
    """
    db_vcs_instances = vcs_instance_crud.get_vcs_instances(db_connection, skip=skip, limit=limit,
                                                           vcs_provider_type=vcs_provider_type,
                                                           vcs_instance_name=vcs_instance_name)
    vcs_instances = []
    for db_vcs_instance in db_vcs_instances:
        vcs_instances.append(vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance))

    total_vcs_instances = vcs_instance_crud.get_vcs_instances_count(db_connection, vcs_provider_type=vcs_provider_type,
                                                                    vcs_instance_name=vcs_instance_name)

    return PaginationModel[vcs_instance_schema.VCSInstanceRead](data=vcs_instances, total=total_vcs_instances,
                                                                limit=limit, skip=skip)


@router.put("/{vcs_instance_id}", response_model=vcs_instance_schema.VCSInstanceRead, status_code=status.HTTP_200_OK)
def update_vcs_instance(vcs_instance_id: int, vcs_instance: vcs_instance_schema.VCSInstanceCreate,
                        db_connection: Session = Depends(get_db_connection)) \
        -> vcs_instance_schema.VCSInstanceRead:
    """
        Update a VCS instance object
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the VCS instance to update
    :param vcs_instance:
        VCSInstanceCreate type object of the VCS Instance to update the values to
    :return: VCSInstanceRead
        The output will contain a VCSInstanceRead type object with the new properties
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS instance not found")
    db_vcs_instance = vcs_instance_crud.update_vcs_instance(db_connection=db_connection,
                                                            vcs_instance_id=vcs_instance_id, vcs_instance=vcs_instance)
    vcs_instance = vcs_instance_schema.VCSInstanceRead.create_from_db_vcs_instance(db_vcs_instance)
    return vcs_instance


@router.delete("/{vcs_instance_id}",
               status_code=status.HTTP_200_OK)
def delete_vcs_instance(vcs_instance_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a VCS instance object
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the VCS instance to delete
    :return:
        The output will contain a success or error message based on the success of the deletion
    """
    db_vcs_instance = vcs_instance_crud.get_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id)
    if db_vcs_instance is None:
        raise HTTPException(status_code=404, detail="VCS instance not found")
    vcs_instance_crud.delete_vcs_instance(db_connection, vcs_instance_id=vcs_instance_id, delete_related=True)
    return {"ok": True}



# Standard Library
import logging
import os
import ssl
import urllib.error

# Third Party
import jwt
import sqlalchemy.orm
from fastapi import Depends, HTTPException, status
from fastapi.security import HTTPBasicCredentials, HTTPBearer
from jwt import PyJWKClient
from tenacity import retry, stop_after_attempt, wait_exponential

# First Party
from resc_backend.constants import RESC_OPERATOR_ROLE
from resc_backend.db.connection import Session, engine
from resc_backend.db.model import DBbranch, DBfinding, DBrepository, DBrule, DBscan, DBscanFinding

security = HTTPBearer()
logger = logging.getLogger(__name__)


async def requires_auth(credentials: HTTPBasicCredentials = Depends(security)):
    """
        Function that is used to validate the JWT access token
    """
    access_token = credentials.credentials
    algorithm = ["RS256"]
    issuer = os.getenv('SSO_ACCESS_TOKEN_ISSUER_URL', '')
    jwks_url = os.getenv('SSO_ACCESS_TOKEN_JWKS_URL', '')
    jwt_options = {
        "verify_signature": True, "verify_iss": True, "verify_exp": True,
        "require": ["client_id", "iss", "corpid", "roles", "given_name", "family_name", "userid", "email",
                    "department_number", "exp"]
    }
    try:
        ssl._create_default_https_context = ssl._create_unverified_context  # pylint: disable=W0212
        jwks_client = PyJWKClient(jwks_url)
        signing_key = jwks_client.get_signing_key_from_jwt(access_token)
        claims = jwt.decode(
            access_token,
            signing_key.key,
            algorithms=algorithm,
            issuer=issuer,
            options=jwt_options,
        )

        if not user_has_resc_operator_role(claims):
            logger.error(f"Invalid login attempt for user {claims['email']}")
            raise HTTPException(status_code=status.HTTP_403_FORBIDDEN,
                                detail="You don't have permission to access this resource.")

    except urllib.error.URLError as error:
        logger.error(f"Unable to contact server for token validation {jwks_url} Message: {error}")
        raise HTTPException(status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                            detail="Unable to contact server for token validation") from error

    except jwt.InvalidAlgorithmError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Algorithm is invalid for decoding token",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.PyJWKClientError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Unable to find the signing key",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.InvalidKeyError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="JWKS key is not in the proper format",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.InvalidIssuerError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail="Token’s issuer claim does not match with the expected issuer",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.ExpiredSignatureError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Token has expired",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.InvalidSignatureError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Token’s signature did not match",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.DecodeError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                            detail="Token can not be decoded because it failed validation",
                            headers={"WWW-Authenticate": "Bearer"}) from error
    except jwt.InvalidTokenError as error:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED, detail="Invalid token",
                            headers={"WWW-Authenticate": "Bearer"}) from error


def user_has_resc_operator_role(claims: dict) -> bool:
    """
        Function that is used to determine if the user has the RESC_OPERATOR_ROLE
    """
    return bool("roles" in claims and claims["roles"] == RESC_OPERATOR_ROLE)


def requires_no_auth():
    """
        Function that is used for unauthenticated access
    """
    return None


def get_db_connection():
    db_connection = Session(bind=engine)
    try:
        yield db_connection
    finally:
        db_connection.close()


@retry(wait=wait_exponential(multiplier=1, min=2, max=10), stop=stop_after_attempt(100))
def check_db_initialized():
    required_table_names = [
        DBbranch.__tablename__,
        DBfinding.__tablename__,
        DBrepository.__tablename__,
        DBrule.__tablename__,
        DBscan.__tablename__,
        DBscanFinding.__tablename__
    ]
    try:
        # Create a sqlalchemy inspector, will cause an exception if the db connection can't be established
        inspector = sqlalchemy.inspect(engine)

        # Check existence of required tables
        not_found_tables = []
        for table_name in required_table_names:
            table_exists = inspector.has_table(table_name)
            if not table_exists:
                not_found_tables.append(table_name)

        if len(not_found_tables) > 0:
            raise Exception(f"Unable to determine existence of required table(s) "
                            f"{', '.join(not_found_tables)}")
    except Exception as ex:
        logger.error(f"Database is NOT connected or initialized | {ex} | Retrying...")
        raise



# Standard Library
import logging
from datetime import datetime
from typing import List, Optional

# Third Party
from fastapi import APIRouter, Depends, Query, status

# First Party
from resc_backend.constants import RULES_TAG, RWS_ROUTE_DETECTED_RULES, RWS_ROUTE_FINDING_STATUS_COUNT, RWS_ROUTE_RULES
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import rule as rule_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.rule import RuleCreate
from resc_backend.resc_web_service.schema.rule_count_model import RuleFindingCountModel
from resc_backend.resc_web_service.schema.status_count import StatusCount
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(tags=[RULES_TAG])

logger = logging.getLogger(__name__)


@router.get(f"{RWS_ROUTE_DETECTED_RULES}",
            response_model=List[str],
            status_code=status.HTTP_200_OK)
def get_distinct_rules_from_findings(
        finding_statuses: List[FindingStatus] = Query(None, alias="findingstatus", title="FindingStatuses"),
        vcs_providers: List[VCSProviders] = Query(None, alias="vcsprovider", title="VCSProviders"),
        project_name: Optional[str] = Query('', regex=r"^[A-z0-9 .\-_%]*$"),
        repository_name: Optional[str] = Query('', regex=r"^[A-z0-9 .\-_%]*$"),
        start_date_time: Optional[datetime] = Query(None),
        end_date_time: Optional[datetime] = Query(None),
        db_connection: Session = Depends(get_db_connection)) -> List[str]:
    """
        Retrieve all uniquely detected rules across all findings in the database
    :param finding_statuses:
        optional, filter of supported finding statuses
    :param vcs_providers:
        optional, filter of supported vcs provider types
    :param db_connection:
        Session of the database connection
    :param project_name:
        optional, filter on project name. Is used as a full string match filter
    :param repository_name:
        Optional, filter on repository name. Is used as a string contains filter
    :param start_date_time
        Optional, filter on start date
    :param end_date_time
        Optional, filter on end date
    :return: List[str]
        The output will contain a list of strings of unique rules in the findings table
    """
    distinct_rules = finding_crud.get_distinct_rules_from_findings(db_connection,
                                                                   finding_statuses=finding_statuses,
                                                                   vcs_providers=vcs_providers,
                                                                   project_name=project_name,
                                                                   repository_name=repository_name,
                                                                   start_date_time=start_date_time,
                                                                   end_date_time=end_date_time)
    rules = [rule.rule_name for rule in distinct_rules]
    return rules


@router.get(f"{RWS_ROUTE_RULES}{RWS_ROUTE_FINDING_STATUS_COUNT}",
            response_model=List[RuleFindingCountModel],
            status_code=status.HTTP_200_OK)
def get_rules_finding_status_count(db_connection: Session = Depends(get_db_connection)) -> List[RuleFindingCountModel]:
    """
        Retrieve all detected rules with finding counts per supported status
    :param db_connection:
        Session of the database connection
    :return: List[str]
        The output will contain a list of strings of unique rules in the findings table
    """
    distinct_rules = finding_crud.get_distinct_rules_from_findings(db_connection)

    rule_findings_counts = []
    for rule in distinct_rules:
        finding_count = 0
        rule_finding_count = RuleFindingCountModel(rule_name=rule.rule_name)
        count_by_status = finding_crud.get_findings_count_by_status(db_connection,
                                                                    rule_name=rule_finding_count.rule_name)
        handled_statuses = []
        for status_count in count_by_status:
            finding_status_count = StatusCount(status=status_count[0], count=status_count[1])
            finding_count = finding_count + finding_status_count.count
            handled_statuses.append(finding_status_count.status)
            rule_finding_count.finding_statuses_count.append(finding_status_count)

        for finding_status in FindingStatus:
            # add default values of 0 for statuses without findings
            if finding_status not in handled_statuses:
                finding_status_count = StatusCount(status=finding_status, count=0)
                rule_finding_count.finding_statuses_count.append(finding_status_count)

        rule_finding_count.finding_count = finding_count
        rule_finding_count.finding_statuses_count = sorted(rule_finding_count.finding_statuses_count,
                                                           key=lambda status_counter: status_counter.status)
        rule_findings_counts.append(rule_finding_count)

    return rule_findings_counts


def create_rule(
        rule: RuleCreate,
        db_connection: Session = Depends(get_db_connection)):
    """
        Create rule in database
    :param rule:
        RuleCreate object to be created
    :param db_connection:
        Session of the database connection
    """
    return rule_crud.create_rule(db_connection=db_connection,
                                 rule=rule)



# Standard Library
from typing import List

# Third Party
from fastapi import APIRouter, Response, status

# First Party
from resc_backend.constants import CACHE_MAX_AGE, COMMON_TAG, RWS_ROUTE_AUTH_CHECK, RWS_ROUTE_SUPPORTED_VCS_PROVIDERS
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders

router = APIRouter(tags=[COMMON_TAG])


@router.get(f"{RWS_ROUTE_SUPPORTED_VCS_PROVIDERS}",
            response_model=List[str],
            status_code=status.HTTP_200_OK)
def get_supported_vcs_providers(response: Response) -> List[str]:
    """
        Retrieve all supported vcs providers
    :return: List[str]
        The output will contain a list of strings of unique rules in the findings table
    """
    response.headers["Cache-Control"] = CACHE_MAX_AGE
    supported_vcs = [vcs for vcs in VCSProviders if vcs]
    return supported_vcs


@router.get(f"{RWS_ROUTE_AUTH_CHECK}",
            status_code=status.HTTP_200_OK)
def auth_check():
    """
        Validates authorization check from the access token
    :return: str
        The output will contain 200 OK if auth check is successful else it will return 403 Forbidden
    """
    return {"message": "OK"}



# Standard Library
from datetime import datetime
from typing import List, Optional

# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, Response, status

# First Party
from resc_backend.constants import (
    CACHE_MAX_AGE,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    FINDINGS_TAG,
    RWS_ROUTE_AUDIT,
    RWS_ROUTE_BY_RULE,
    RWS_ROUTE_COUNT_BY_TIME,
    RWS_ROUTE_FINDINGS,
    RWS_ROUTE_SUPPORTED_STATUSES,
    RWS_ROUTE_TOTAL_COUNT_BY_RULE
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.filters import FindingsFilter
from resc_backend.resc_web_service.schema import audit as audit_schema
from resc_backend.resc_web_service.schema import finding as finding_schema
from resc_backend.resc_web_service.schema.date_count_model import DateCountModel
from resc_backend.resc_web_service.schema.date_filter import DateFilter
from resc_backend.resc_web_service.schema.finding import FindingRead
from resc_backend.resc_web_service.schema.finding_status import FindingStatus
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel

router = APIRouter(prefix=f"{RWS_ROUTE_FINDINGS}", tags=[FINDINGS_TAG])


@router.get("",
            response_model=PaginationModel[finding_schema.FindingRead],
            status_code=status.HTTP_200_OK)
def get_all_findings(skip: int = Query(default=0, ge=0),
                     limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                     db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all findings objects paginated
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no finding was found
    """
    findings = finding_crud.get_findings(db_connection, skip=skip, limit=limit)

    total_findings = finding_crud.get_total_findings_count(db_connection)

    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.post("",
             response_model=int,
             status_code=status.HTTP_201_CREATED)
def create_findings(findings: List[finding_schema.FindingCreate], db_connection: Session = Depends(get_db_connection)) \
        -> int:
    """
          Create new findings
      :param db_connection:
          Session of the database connection
      :param findings:
          List of findings to create
      :return: int
          The output will contain the number of successful created findings
      """
    try:
        created_findings = finding_crud.create_findings(db_connection=db_connection, findings=findings)

    except KeyError as err:
        raise HTTPException(status_code=400, detail=str(err)) from err
    return len(created_findings)


@router.get("/{finding_id}",
            response_model=finding_schema.FindingRead,
            status_code=status.HTTP_200_OK)
def read_finding(finding_id: int, db_connection: Session = Depends(get_db_connection)):
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    db_scan_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
    scan_ids = [x.scan_id for x in db_scan_findings]
    return FindingRead.create_from_db_entities(db_finding=db_finding, scan_ids=scan_ids)


@router.patch("/{finding_id}",
              response_model=finding_schema.FindingRead,
              status_code=status.HTTP_200_OK)
def patch_finding(
        finding_id: int,
        finding_update: finding_schema.FindingPatch,
        db_connection: Session = Depends(get_db_connection)
):
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    db_sca_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
    scan_ids = [x.scan_id for x in db_sca_findings]
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    return FindingRead.create_from_db_entities(
        finding_crud.patch_finding(db_connection, finding_id=finding_id, finding_update=finding_update),
        scan_ids
    )


@router.put("/{finding_id}",
            response_model=finding_schema.FindingRead,
            status_code=status.HTTP_200_OK)
def update_finding(
        finding_id: int,
        finding: finding_schema.FindingUpdate,
        db_connection: Session = Depends(get_db_connection)
):
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    db_sca_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
    scan_ids = [x.scan_id for x in db_sca_findings]
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    return FindingRead.create_from_db_entities(
        finding_crud.update_finding(db_connection=db_connection, finding_id=finding_id, finding=finding),
        scan_ids
    )


@router.delete("/{finding_id}",
               status_code=status.HTTP_200_OK)
def delete_finding(finding_id: int, db_connection: Session = Depends(get_db_connection)) -> FindingRead:
    """
        Delete a finding object
    :param db_connection:
        Session of the database connection
    :param finding_id:
        id of the finding to delete
    :return:
        The output will contain a success or error message based on the success of the deletion
    """
    db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
    if db_finding is None:
        raise HTTPException(status_code=404, detail="Finding not found")
    finding_crud.delete_finding(db_connection, finding_id=finding_id, delete_related=True)
    return {"ok": True}


@router.get(f"{RWS_ROUTE_TOTAL_COUNT_BY_RULE}""/{rule_name}",
            status_code=status.HTTP_200_OK)
def get_total_findings_count_by_rule(rule_name: str, db_connection: Session = Depends(get_db_connection)):
    findings_filter = FindingsFilter(rule_names=[rule_name])
    return finding_crud.get_total_findings_count(db_connection, findings_filter=findings_filter)


@router.get(f"{RWS_ROUTE_BY_RULE}""/{rule_name}",
            response_model=PaginationModel[finding_schema.FindingRead],
            status_code=status.HTTP_200_OK)
def get_findings_by_rule(rule_name: str, skip: int = Query(default=0, ge=0),
                         limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                         db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[finding_schema.FindingRead]:
    """
        Retrieve all findings objects paginated by rule
    :param db_connection:
        Session of the database connection
    :param rule_name:
        Name of the rule to filter the findings by
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [FindingRead]
        The output will contain a PaginationModel containing the list of FindingRead type objects,
        or an empty list if no finding was found for the given rule
    """
    findings = finding_crud.get_findings_by_rule(db_connection, skip=skip, limit=limit, rule_name=rule_name)
    total_findings = finding_crud.get_total_findings_count(
        db_connection, findings_filter=FindingsFilter(rule_names=[rule_name]))
    return PaginationModel[finding_schema.FindingRead](data=findings, total=total_findings, limit=limit, skip=skip)


@router.put(f"{RWS_ROUTE_AUDIT}/",
            response_model=List[finding_schema.FindingRead],
            status_code=status.HTTP_200_OK)
def audit_findings(
        audit: audit_schema.AuditMultiple,
        db_connection: Session = Depends(get_db_connection)
) -> List[finding_schema.FindingRead]:
    """
        Audit single/multiple findings, updating the status and comment
    :param db_connection:
        Session of the database connection
    :param audit:
        audit object, containing the ids to audit, status to set and the comment
    :return: [FindingRead]
        The output will contain a list of the findings that where updated
    """
    audited_findings = []
    for finding_id in audit.finding_ids:
        db_finding = finding_crud.get_finding(db_connection, finding_id=finding_id)
        db_scan_findings = scan_finding_crud.get_scan_findings(db_connection, finding_id=finding_id)
        scan_ids = [x.scan_id for x in db_scan_findings]
        if db_finding is None:
            raise HTTPException(status_code=404, detail=f"Finding {finding_id} not found")

        audited_findings.append(
            FindingRead.create_from_db_entities(
                finding_crud.audit_finding(db_connection=db_connection, db_finding=db_finding,
                                           status=audit.status, comment=audit.comment),
                scan_ids=scan_ids)
        )
    return audited_findings


@router.get(f"{RWS_ROUTE_SUPPORTED_STATUSES}/",
            response_model=List[str],
            status_code=status.HTTP_200_OK)
def get_supported_statuses(response: Response) -> List[str]:
    """
        Retrieve all supported statuses for findings
    :return: List[str]
        The output will contain a list of strings of unique statuses supported
    """
    response.headers["Cache-Control"] = CACHE_MAX_AGE
    supported_finding_statuses = [finding_status for finding_status in FindingStatus if finding_status]
    return supported_finding_statuses


@router.get(f"{RWS_ROUTE_COUNT_BY_TIME}/""{time_type}",
            response_model=PaginationModel[DateCountModel],
            status_code=status.HTTP_200_OK)
def get_count_by_time(time_type: DateFilter,
                      skip: int = Query(default=0, ge=0),
                      limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                      start_date_time: Optional[datetime] = Query(None),
                      end_date_time: Optional[datetime] = Query(None),
                      db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[DateCountModel]:
    """
        Retrieve all findings count by time period objects paginated
    :param db_connection:
        Session of the database connection
    :param time_type:
        required, filter on time_type
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :param start_date_time
        Optional, filter on start date
    :param end_date_time
        Optional, filter on end date
    :return: PaginationModel[DateCountModel]
        The output will contain a PaginationModel containing the list of DateCountModel type objects,
        or an empty list if no data was found
    """
    date_counts = []
    findings = finding_crud.get_findings_count_by_time(db_connection=db_connection, date_type=time_type,
                                                       start_date_time=start_date_time, end_date_time=end_date_time,
                                                       skip=skip, limit=limit)
    total_findings = finding_crud.get_findings_count_by_time_total(db_connection=db_connection, date_type=time_type,
                                                                   start_date_time=start_date_time,
                                                                   end_date_time=end_date_time)
    for finding in findings:
        if time_type == DateFilter.MONTH:
            date_count = DateCountModel(finding_count=finding[2], date_lable=f"{finding[0]}-{finding[1]}")
        elif time_type == DateFilter.WEEK:
            date_count = DateCountModel(finding_count=finding[2], date_lable=f"{finding[0]}-W{finding[1]}")
        elif time_type == DateFilter.DAY:
            date_count = DateCountModel(finding_count=finding[3], date_lable=f"{finding[0]}-{finding[1]}-{finding[2]}")

        date_counts.append(date_count)

    return PaginationModel[DateCountModel](data=date_counts, total=total_findings, limit=limit, skip=skip)



# Standard Library
import logging

# Third Party
from fastapi import APIRouter, status

# First Party
from resc_backend.constants import HEALTH_TAG, RWS_ROUTE_HEALTH

router = APIRouter(tags=[HEALTH_TAG])

logger = logging.getLogger(__name__)


@router.get(f"{RWS_ROUTE_HEALTH}",
            status_code=status.HTTP_200_OK)
def health_check():
    return {"status": "OK"}



# Third Party
from fastapi import APIRouter, Depends, HTTPException, Query, status

# First Party
from resc_backend.constants import (
    BRANCHES_TAG,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    RWS_ROUTE_BRANCHES,
    RWS_ROUTE_FINDINGS_METADATA,
    RWS_ROUTE_LAST_SCAN,
    RWS_ROUTE_SCANS
)
from resc_backend.db.connection import Session
from resc_backend.resc_web_service.crud import branch as branch_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.schema import branch as branch_schema
from resc_backend.resc_web_service.schema import scan as scan_schema
from resc_backend.resc_web_service.schema.finding_count_model import FindingCountModel
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel

router = APIRouter(prefix=f"{RWS_ROUTE_BRANCHES}", tags=[BRANCHES_TAG])


@router.get("",
            response_model=PaginationModel[branch_schema.BranchRead],
            status_code=status.HTTP_200_OK)
def get_all_branches(skip: int = Query(default=0, ge=0),
                     limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                     db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[branch_schema.BranchRead]:
    """
        Retrieve all branch objects paginated
    :param db_connection:
        Session of the database connection
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [BranchRead]
        The output will contain a PaginationModel containing the list of BranchRead type objects,
        or an empty list if no branch was found
    """
    branches = branch_crud.get_branches(db_connection, skip=skip, limit=limit)

    total_branches = branch_crud.get_branches_count(db_connection)

    return PaginationModel[branch_schema.BranchRead](data=branches, total=total_branches,
                                                     limit=limit, skip=skip)


@router.post("",
             response_model=branch_schema.BranchRead, status_code=status.HTTP_201_CREATED)
def create_branch(
        branch: branch_schema.BranchCreate, db_connection: Session = Depends(get_db_connection)):
    return branch_crud.create_branch_if_not_exists(db_connection=db_connection, branch=branch)


@router.get("/{branch_id}",
            response_model=branch_schema.BranchRead,
            status_code=status.HTTP_200_OK)
def read_branch(branch_id: int, db_connection: Session = Depends(get_db_connection)):
    db_branch = branch_crud.get_branch(db_connection, branch_id=branch_id)
    if db_branch is None:
        raise HTTPException(status_code=404, detail="Branch not found")
    return db_branch


@router.put("/{branch_id}",
            response_model=branch_schema.BranchRead,
            status_code=status.HTTP_200_OK)
def update_branch(
        branch_id: int,
        branch: branch_schema.BranchCreate,
        db_connection: Session = Depends(get_db_connection)
):
    db_branch = branch_crud.get_branch(db_connection, branch_id=branch_id)
    if db_branch is None:
        raise HTTPException(status_code=404, detail="Branch not found")
    return branch_crud.update_branch(
        db_connection=db_connection,
        branch_id=branch_id,
        branch=branch
    )


@router.delete("/{branch_id}",
               status_code=status.HTTP_200_OK)
def delete_branch(branch_id: int, db_connection: Session = Depends(get_db_connection)):
    """
        Delete a branch object
    :param db_connection:
        Session of the database connection
    :param branch_id:
        id of the branch to delete
    :return:
        The output will contain a success or error message based on the success of the deletion
    """
    db_branch = branch_crud.get_branch(db_connection, branch_id=branch_id)
    if db_branch is None:
        raise HTTPException(status_code=404, detail="Branch not found")
    branch_crud.delete_branch(db_connection, branch_id=branch_id, delete_related=True)
    return {"ok": True}


@router.get("/{branch_id}"f"{RWS_ROUTE_SCANS}",
            response_model=PaginationModel[scan_schema.ScanRead],
            status_code=status.HTTP_200_OK)
def get_scans_for_branch(branch_id: int, skip: int = Query(default=0, ge=0),
                         limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
                         db_connection: Session = Depends(get_db_connection)) \
        -> PaginationModel[scan_schema.ScanRead]:
    """
        Retrieve all scan objects related to a branch paginated
    :param db_connection:
        Session of the database connection
    :param branch_id:
        id of the parent branch object of which to retrieve scan objects
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [ScanRead]
        The output will contain a PaginationModel containing the list of ScanRead type objects,
        or an empty list if no scan was found
    """
    scans = scan_crud.get_scans(db_connection, skip=skip, limit=limit, branch_id=branch_id)

    total_scans = scan_crud.get_scans_count(db_connection, branch_id=branch_id)

    return PaginationModel[scan_schema.ScanRead](data=scans, total=total_scans, limit=limit, skip=skip)


@router.get("/{branch_id}"f"{RWS_ROUTE_LAST_SCAN}",
            response_model=scan_schema.ScanRead,
            status_code=status.HTTP_200_OK)
def get_last_scan_for_branch(branch_id: int, db_connection: Session = Depends(get_db_connection)) \
        -> scan_schema.ScanRead:
    """
        Retrieve latest scan object related to a branch
    :param db_connection:
        Session of the database connection
    :param branch_id:
        id of the parent branch object of which to retrieve scan object
    :return: ScanRead
        The output will contain a ScanRead type object,
        or empty if no scan was found
    """
    last_scan = scan_crud.get_latest_scan_for_branch(db_connection, branch_id=branch_id)

    return last_scan


@router.get("/{branch_id}"f"{RWS_ROUTE_FINDINGS_METADATA}",
            response_model=FindingCountModel[branch_schema.BranchRead],
            status_code=status.HTTP_200_OK)
def get_findings_metadata_for_branch(branch_id: int,
                                     db_connection: Session = Depends(get_db_connection)) \
        -> FindingCountModel[branch_schema.BranchRead]:
    """
        Retrieve findings metadata for a branch
    :param branch_id:
        id of the branch object for which findings metadata to be retrieved
    :param db_connection:
        Session of the database connection
    :return: BranchedRead, findings count per status
        The output will contain a BranchedRead type object along with findings count per status,
        or empty if no scan was found
    """
    branch = branch_crud.get_branch(db_connection, branch_id=branch_id)
    if branch is None:
        raise HTTPException(status_code=404, detail="Branch not found")

    findings_meta_data = branch_crud.get_findings_metadata_by_branch_id(db_connection,
                                                                        branch_id=branch_id)

    return FindingCountModel[branch_schema.BranchRead](
        data=branch,
        true_positive=findings_meta_data["true_positive"],
        false_positive=findings_meta_data["false_positive"],
        not_analyzed=findings_meta_data["not_analyzed"],
        under_review=findings_meta_data["under_review"],
        clarification_required=findings_meta_data["clarification_required"],
        total_findings_count=findings_meta_data["total_findings_count"])



# Third Party
from sqlalchemy import func
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import DEFAULT_RECORDS_PER_PAGE_LIMIT, MAX_RECORDS_PER_PAGE_LIMIT
from resc_backend.db import model
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import branch as branch_schema


def get_branches(db_connection: Session, skip: int = 0,
                 limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT):
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    branches = db_connection.query(model.DBbranch).order_by(
        model.branch.DBbranch.id_).offset(skip).limit(limit_val).all()
    return branches


def get_branches_for_repository(db_connection: Session, repository_id: int, skip: int = 0,
                                limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT) -> [model.DBbranch]:
    """
        Retrieve all branch child objects of a repository object from the database
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the parent repository object of which to retrieve branch objects
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBbranch]
        The output will contain a list of DBbranch type objects,
        or an empty list if no branch was found for the given repository_id
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    branches = db_connection.query(model.DBbranch) \
        .filter(model.DBbranch.repository_id == repository_id) \
        .order_by(model.branch.DBbranch.id_).offset(skip).limit(limit_val).all()
    return branches


def get_branches_count(db_connection: Session) -> int:
    """
        Retrieve count of branches records
    :param db_connection:
        Session of the database connection
    :return: total_count
        count of branches
    """
    total_count = db_connection.query(func.count(model.DBbranch.id_)).scalar()
    return total_count


def get_branches_count_for_repository(db_connection: Session, repository_id: int) -> int:
    """
        Retrieve count of finding records of a given scan
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository_id object for which to retrieve the count of branches
    :return: total_count
        count of branches
    """
    total_count = db_connection.query(func.count(model.DBbranch.id_)) \
        .filter(model.DBbranch.repository_id == repository_id) \
        .scalar()
    return total_count


def get_branch(db_connection: Session, branch_id: int):
    branch = db_connection.query(model.DBbranch) \
        .filter(model.branch.DBbranch.id_ == branch_id).first()
    return branch


def update_branch(db_connection: Session, branch_id: int, branch: branch_schema.BranchCreate):
    db_branch = db_connection.query(model.DBbranch).filter_by(id_=branch_id).first()

    db_branch.branch_name = branch.branch_name
    db_branch.latest_commit = branch.latest_commit

    db_connection.commit()
    db_connection.refresh(db_branch)
    return db_branch


def create_branch(db_connection: Session, branch: branch_schema.BranchCreate):
    db_branch = model.branch.DBbranch(
        repository_id=branch.repository_id,
        branch_id=branch.branch_id,
        branch_name=branch.branch_name,
        latest_commit=branch.latest_commit
    )
    db_connection.add(db_branch)
    db_connection.commit()
    db_connection.refresh(db_branch)
    return db_branch


def create_branch_if_not_exists(db_connection: Session, branch: branch_schema.BranchCreate):
    # Query the database to see if the branch object exists based on the unique constraint parameters
    db_select_branch = db_connection.query(model.DBbranch) \
        .filter(model.DBbranch.branch_id == branch.branch_id,
                model.DBbranch.repository_id == branch.repository_id).first()
    if db_select_branch is not None:
        return db_select_branch

    # Create non-existing branch object
    return create_branch(db_connection, branch)


def get_findings_metadata_by_branch_id(db_connection: Session, branch_id: int):
    """
        Retrieves the finding metadata for a branch id from the database with most recent scan information
    :param db_connection:
        Session of the database connection
    :param branch_id:
        id of the branch for which findings metadata to be retrieved
    :return: findings_metadata
        findings_metadata containing the count for each status
    """

    latest_scan = scan_crud.get_latest_scan_for_branch(db_connection, branch_id=branch_id)

    if latest_scan is not None:
        findings_metadata = scan_crud.get_branch_findings_metadata_for_latest_scan(
            db_connection, branch_id=latest_scan.branch_id, scan_timestamp=latest_scan.timestamp)
    else:
        findings_metadata = {
            "true_positive": 0,
            "false_positive": 0,
            "not_analyzed": 0,
            "under_review": 0,
            "clarification_required": 0,
            "total_findings_count": 0
        }

    return findings_metadata


def delete_branch(db_connection: Session, branch_id: int, delete_related: bool = False):
    """
        Delete a branch object
    :param db_connection:
        Session of the database connection
    :param branch_id:
        id of the branch to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding_by_branch_id(db_connection, branch_id=branch_id)
        finding_crud.delete_findings_by_branch_id(db_connection, branch_id=branch_id)
        scan_crud.delete_scans_by_branch_id(db_connection, branch_id=branch_id)
    db_connection.query(model.DBbranch) \
        .filter(model.branch.DBbranch.id_ == branch_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_branches_by_repository_id(db_connection: Session, repository_id: int):
    """
        Delete branches for a given repository
    :param db_connection:
        Session of the database connection
    :param repository_id:
        id of the repository
    """
    db_connection.query(model.DBbranch) \
        .filter(model.repository.DBrepository.id_ == model.branch.DBbranch.repository_id,
                model.repository.DBrepository.id_ == repository_id) \
        .delete(synchronize_session=False)
    db_connection.commit()


def delete_branches_by_vcs_instance_id(db_connection: Session, vcs_instance_id: int):
    """
        Delete branches for a given vcs instance
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance
    """
    db_connection.query(model.DBbranch) \
        .filter(model.repository.DBrepository.id_ == model.branch.DBbranch.repository_id,
                model.repository.DBrepository.vcs_instance == model.vcs_instance.DBVcsInstance.id_,
                model.vcs_instance.DBVcsInstance.id_ == vcs_instance_id) \
        .delete(synchronize_session=False)
    db_connection.commit()



# Third Party
from sqlalchemy import Column, ForeignKey, Integer, String, UniqueConstraint

# First Party
from resc_backend.db.model import Base
from resc_backend.resc_web_service.schema.branch import Branch


class DBbranch(Base):
    __tablename__ = "branch"
    id_ = Column("id", Integer, primary_key=True)
    repository_id = Column(Integer, ForeignKey("repository.id"), nullable=False)
    branch_id = Column(String(200), nullable=False)
    branch_name = Column(String(200), nullable=False)
    latest_commit = Column(String(100), nullable=False)
    __table_args__ = (UniqueConstraint("branch_id", "repository_id", name="unique_branch_id_per_repository"),)

    def __init__(self, repository_id, branch_id, branch_name, latest_commit):
        self.branch_id = branch_id
        self.repository_id = repository_id
        self.branch_name = branch_name
        self.latest_commit = latest_commit

    @staticmethod
    def create_from_branch(branch: Branch, repository_id: int):
        db_branch = DBbranch(
            repository_id=repository_id,
            branch_id=branch.branch_id,
            branch_name=branch.branch_name,
            latest_commit=branch.latest_commit
        )
        return db_branch
